/*
 * wpa_supplicant - Event notifications
 * Copyright (c) 2009-2010, Jouni Malinen <j@w1.fi>
 *
 * This software may be distributed under the terms of the BSD license.
 * See README for more details.
 */

#include "utils/includes.h"

#include "utils/common.h"
#include "common/wpa_ctrl.h"
#include "common/nan_de.h"
#include "common/proximity_ranging.h"
#include "config.h"
#include "wpa_supplicant_i.h"
#include "wps_supplicant.h"
#include "binder/binder.h"
#include "dbus/dbus_common.h"
#include "dbus/dbus_new.h"
#include "rsn_supp/wpa.h"
#include "rsn_supp/pmksa_cache.h"
#include "fst/fst.h"
#include "crypto/tls.h"
#include "bss.h"
#include "driver_i.h"
#include "scan.h"
#include "p2p_supplicant.h"
#include "sme.h"
#include "notify.h"

int wpas_notify_supplicant_initialized(struct wpa_global *global)
{
#ifdef CONFIG_CTRL_IFACE_DBUS_NEW
	if (global->params.dbus_ctrl_interface) {
		global->dbus = wpas_dbus_init(global);
		if (global->dbus == NULL)
			return -1;
	}
#endif /* CONFIG_CTRL_IFACE_DBUS_NEW */

#ifdef CONFIG_BINDER
	global->binder = wpas_binder_init(global);
	if (!global->binder)
		return -1;
#endif /* CONFIG_BINDER */

	return 0;
}


void wpas_notify_supplicant_deinitialized(struct wpa_global *global)
{
#ifdef CONFIG_CTRL_IFACE_DBUS_NEW
	if (global->dbus)
		wpas_dbus_deinit(global->dbus);
#endif /* CONFIG_CTRL_IFACE_DBUS_NEW */

#ifdef CONFIG_BINDER
	if (global->binder)
		wpas_binder_deinit(global->binder);
#endif /* CONFIG_BINDER */
}


int wpas_notify_iface_added(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return 0;

	if (wpas_dbus_register_interface(wpa_s))
		return -1;

	return 0;
}


void wpas_notify_iface_removed(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	/* unregister interface in new DBus ctrl iface */
	wpas_dbus_unregister_interface(wpa_s);
}


void wpas_notify_state_changed(struct wpa_supplicant *wpa_s,
			       enum wpa_states new_state,
			       enum wpa_states old_state)
{
	struct wpa_ssid *ssid = wpa_s->current_ssid;

	if (wpa_s->p2p_mgmt)
		return;

	/* notify the new DBus API */
	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_STATE);

#ifdef CONFIG_FST
	if (wpa_s->fst && !is_zero_ether_addr(wpa_s->bssid)) {
		if (new_state == WPA_COMPLETED)
			fst_notify_peer_connected(wpa_s->fst, wpa_s->bssid);
		else if (old_state >= WPA_ASSOCIATED &&
			 new_state < WPA_ASSOCIATED)
			fst_notify_peer_disconnected(wpa_s->fst, wpa_s->bssid);
	}
#endif /* CONFIG_FST */

	if (new_state == WPA_COMPLETED) {
		wpas_p2p_notif_connected(wpa_s);
		if (ssid && !wpa_s->sta_roaming_disabled)
			wpa_drv_roaming(wpa_s, !ssid->bssid_set,
					ssid->bssid_set ? ssid->bssid : NULL);
	} else if (old_state >= WPA_ASSOCIATED && new_state < WPA_ASSOCIATED) {
		wpas_p2p_notif_disconnected(wpa_s);
	}

	sme_state_changed(wpa_s);

#ifdef ANDROID
	wpa_msg_ctrl(wpa_s, MSG_INFO, WPA_EVENT_STATE_CHANGE
		     "id=%d state=%d BSSID=" MACSTR " SSID=%s",
		     wpa_s->current_ssid ? wpa_s->current_ssid->id : -1,
		     new_state,
		     MAC2STR(wpa_s->bssid),
		     wpa_s->current_ssid && wpa_s->current_ssid->ssid ?
		     wpa_ssid_txt(wpa_s->current_ssid->ssid,
				  wpa_s->current_ssid->ssid_len) : "");
#endif /* ANDROID */
}


void wpas_notify_disconnect_reason(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_DISCONNECT_REASON);
}


void wpas_notify_auth_status_code(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_AUTH_STATUS_CODE);
}


void wpas_notify_assoc_status_code(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_ASSOC_STATUS_CODE);
}


void wpas_notify_roam_time(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_ROAM_TIME);
}


void wpas_notify_roam_complete(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_ROAM_COMPLETE);
}


void wpas_notify_scan_in_progress_6ghz(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s,
				      WPAS_DBUS_PROP_SCAN_IN_PROGRESS_6GHZ);
}


void wpas_notify_session_length(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_SESSION_LENGTH);
}


void wpas_notify_bss_tm_status(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_BSS_TM_STATUS);
}


void wpas_notify_network_changed(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_CURRENT_NETWORK);
}


void wpas_notify_ap_scan_changed(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_AP_SCAN);
}


void wpas_notify_bssid_changed(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_CURRENT_BSS);
}


void wpas_notify_mac_address_changed(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_MAC_ADDRESS);
}


void wpas_notify_auth_changed(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_CURRENT_AUTH_MODE);
}


void wpas_notify_network_enabled_changed(struct wpa_supplicant *wpa_s,
					 struct wpa_ssid *ssid)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_network_enabled_changed(wpa_s, ssid);
}


void wpas_notify_network_selected(struct wpa_supplicant *wpa_s,
				  struct wpa_ssid *ssid)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_network_selected(wpa_s, ssid->id);
}


void wpas_notify_network_request(struct wpa_supplicant *wpa_s,
				 struct wpa_ssid *ssid,
				 enum wpa_ctrl_req_type rtype,
				 const char *default_txt)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_network_request(wpa_s, ssid, rtype, default_txt);
}


void wpas_notify_scanning(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	/* notify the new DBus API */
	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_SCANNING);
}


void wpas_notify_scan_done(struct wpa_supplicant *wpa_s, int success)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_scan_done(wpa_s, success);
}


void wpas_notify_scan_results(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_wps_notify_scan_results(wpa_s);
}


void wpas_notify_wps_credential(struct wpa_supplicant *wpa_s,
				const struct wps_credential *cred)
{
	if (wpa_s->p2p_mgmt)
		return;

#ifdef CONFIG_WPS
	/* notify the new DBus API */
	wpas_dbus_signal_wps_cred(wpa_s, cred);
#endif /* CONFIG_WPS */
}


void wpas_notify_wps_event_m2d(struct wpa_supplicant *wpa_s,
			       struct wps_event_m2d *m2d)
{
	if (wpa_s->p2p_mgmt)
		return;

#ifdef CONFIG_WPS
	wpas_dbus_signal_wps_event_m2d(wpa_s, m2d);
#endif /* CONFIG_WPS */
}


void wpas_notify_wps_event_fail(struct wpa_supplicant *wpa_s,
				struct wps_event_fail *fail)
{
	if (wpa_s->p2p_mgmt)
		return;

#ifdef CONFIG_WPS
	wpas_dbus_signal_wps_event_fail(wpa_s, fail);
#endif /* CONFIG_WPS */
}


void wpas_notify_wps_event_success(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

#ifdef CONFIG_WPS
	wpas_dbus_signal_wps_event_success(wpa_s);
#endif /* CONFIG_WPS */
}

void wpas_notify_wps_event_pbc_overlap(struct wpa_supplicant *wpa_s)
{
	if (wpa_s->p2p_mgmt)
		return;

#ifdef CONFIG_WPS
	wpas_dbus_signal_wps_event_pbc_overlap(wpa_s);
#endif /* CONFIG_WPS */
}


void wpas_notify_network_added(struct wpa_supplicant *wpa_s,
			       struct wpa_ssid *ssid)
{
	if (wpa_s->p2p_mgmt)
		return;

	/*
	 * Networks objects created during any P2P activities should not be
	 * exposed out. They might/will confuse certain non-P2P aware
	 * applications since these network objects won't behave like
	 * regular ones.
	 */
	if (!ssid->p2p_group && wpa_s->global->p2p_group_formation != wpa_s) {
		wpas_dbus_register_network(wpa_s, ssid);
		wpa_msg_ctrl(wpa_s, MSG_INFO, WPA_EVENT_NETWORK_ADDED "%d",
			     ssid->id);
	}
}


void wpas_notify_persistent_group_added(struct wpa_supplicant *wpa_s,
					struct wpa_ssid *ssid)
{
#ifdef CONFIG_P2P
	wpas_dbus_register_persistent_group(wpa_s, ssid);
#endif /* CONFIG_P2P */
}


void wpas_notify_persistent_group_removed(struct wpa_supplicant *wpa_s,
					  struct wpa_ssid *ssid)
{
#ifdef CONFIG_P2P
	wpas_dbus_unregister_persistent_group(wpa_s, ssid->id);
#endif /* CONFIG_P2P */
}


void wpas_notify_network_removed(struct wpa_supplicant *wpa_s,
				 struct wpa_ssid *ssid)
{
	if (wpa_s->next_ssid == ssid)
		wpa_s->next_ssid = NULL;
	if (wpa_s->last_ssid == ssid)
		wpa_s->last_ssid = NULL;
	if (wpa_s->current_ssid == ssid)
		wpa_s->current_ssid = NULL;
	if (wpa_s->ml_connect_probe_ssid == ssid) {
		wpa_s->ml_connect_probe_ssid = NULL;
		wpa_s->ml_connect_probe_bss = NULL;
	}
	if (wpa_s->connect_without_scan == ssid)
		wpa_s->connect_without_scan = NULL;
#if defined(CONFIG_SME) && defined(CONFIG_SAE)
	if (wpa_s->sme.ext_auth_wpa_ssid == ssid)
		wpa_s->sme.ext_auth_wpa_ssid = NULL;
#endif /* CONFIG_SME && CONFIG_SAE */
	if (wpa_s->wpa) {
		if ((wpa_key_mgmt_sae(ssid->key_mgmt) &&
		     (wpa_s->drv_flags2 & WPA_DRIVER_FLAGS2_SAE_OFFLOAD_STA)) ||
		    ((ssid->key_mgmt & WPA_KEY_MGMT_OWE) &&
		     (wpa_s->drv_flags2 & WPA_DRIVER_FLAGS2_OWE_OFFLOAD_STA))) {
			/* For cases when PMK is generated at the driver */
			struct wpa_pmkid_params params;

			os_memset(&params, 0, sizeof(params));
			params.ssid = ssid->ssid;
			params.ssid_len = ssid->ssid_len;
			wpa_drv_remove_pmkid(wpa_s, &params);
		}
		wpa_sm_pmksa_cache_flush(wpa_s->wpa, ssid);
	}
	if (!ssid->p2p_group && wpa_s->global->p2p_group_formation != wpa_s &&
	    !wpa_s->p2p_mgmt) {
		wpas_dbus_unregister_network(wpa_s, ssid->id);
		wpa_msg_ctrl(wpa_s, MSG_INFO, WPA_EVENT_NETWORK_REMOVED "%d",
			     ssid->id);
	}
	if (network_is_persistent_group(ssid))
		wpas_notify_persistent_group_removed(wpa_s, ssid);

	wpas_p2p_network_removed(wpa_s, ssid);
}


void wpas_notify_bss_added(struct wpa_supplicant *wpa_s,
			   u8 bssid[], unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_register_bss(wpa_s, bssid, id);
	wpa_msg_ctrl(wpa_s, MSG_INFO, WPA_EVENT_BSS_ADDED "%u " MACSTR,
		     id, MAC2STR(bssid));
}


void wpas_notify_bss_removed(struct wpa_supplicant *wpa_s,
			     u8 bssid[], unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_unregister_bss(wpa_s, bssid, id);
	wpa_msg_ctrl(wpa_s, MSG_INFO, WPA_EVENT_BSS_REMOVED "%u " MACSTR,
		     id, MAC2STR(bssid));
}


void wpas_notify_bss_freq_changed(struct wpa_supplicant *wpa_s,
				  unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_FREQ, id);
}


void wpas_notify_bss_signal_changed(struct wpa_supplicant *wpa_s,
				    unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_SIGNAL,
					  id);
}


void wpas_notify_bss_privacy_changed(struct wpa_supplicant *wpa_s,
				     unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_PRIVACY,
					  id);
}


void wpas_notify_bss_mode_changed(struct wpa_supplicant *wpa_s,
				  unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_MODE, id);
}


void wpas_notify_bss_wpaie_changed(struct wpa_supplicant *wpa_s,
				   unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_WPA, id);
}


void wpas_notify_bss_rsnie_changed(struct wpa_supplicant *wpa_s,
				   unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_RSN, id);
}


void wpas_notify_bss_wps_changed(struct wpa_supplicant *wpa_s,
				 unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

#ifdef CONFIG_WPS
	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_WPS, id);
#endif /* CONFIG_WPS */
}


void wpas_notify_bss_ies_changed(struct wpa_supplicant *wpa_s,
				   unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_IES, id);
}


void wpas_notify_bss_rates_changed(struct wpa_supplicant *wpa_s,
				   unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_RATES, id);
}


void wpas_notify_bss_seen(struct wpa_supplicant *wpa_s, unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_AGE, id);
}


void wpas_notify_bss_anqp_changed(struct wpa_supplicant *wpa_s, unsigned int id)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_bss_signal_prop_changed(wpa_s, WPAS_DBUS_BSS_PROP_ANQP, id);
}


void wpas_notify_blob_added(struct wpa_supplicant *wpa_s, const char *name)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_blob_added(wpa_s, name);
}


void wpas_notify_blob_removed(struct wpa_supplicant *wpa_s, const char *name)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_blob_removed(wpa_s, name);
}


void wpas_notify_debug_level_changed(struct wpa_global *global)
{
	wpas_dbus_signal_debug_level_changed(global);
}


void wpas_notify_debug_timestamp_changed(struct wpa_global *global)
{
	wpas_dbus_signal_debug_timestamp_changed(global);
}


void wpas_notify_debug_show_keys_changed(struct wpa_global *global)
{
	wpas_dbus_signal_debug_show_keys_changed(global);
}


void wpas_notify_suspend(struct wpa_global *global)
{
	struct wpa_supplicant *wpa_s;

	os_get_time(&global->suspend_time);
	wpa_printf(MSG_DEBUG, "System suspend notification");
	for (wpa_s = global->ifaces; wpa_s; wpa_s = wpa_s->next)
		wpa_drv_suspend(wpa_s);
}


void wpas_notify_resume(struct wpa_global *global)
{
	struct os_time now;
	int slept;
	struct wpa_supplicant *wpa_s;

	if (global->suspend_time.sec == 0)
		slept = -1;
	else {
		os_get_time(&now);
		slept = now.sec - global->suspend_time.sec;
	}
	wpa_printf(MSG_DEBUG, "System resume notification (slept %d seconds)",
		   slept);

	for (wpa_s = global->ifaces; wpa_s; wpa_s = wpa_s->next) {
		wpa_drv_resume(wpa_s);
		if (wpa_s->wpa_state == WPA_DISCONNECTED)
			wpa_supplicant_req_scan(wpa_s, 0, 100000);
	}
}


#ifdef CONFIG_P2P

void wpas_notify_p2p_find_stopped(struct wpa_supplicant *wpa_s)
{
	/* Notify P2P find has stopped */
	wpas_dbus_signal_p2p_find_stopped(wpa_s);
}


void wpas_notify_p2p_device_found(struct wpa_supplicant *wpa_s,
				  const u8 *dev_addr, int new_device)
{
	if (new_device) {
		/* Create the new peer object */
		wpas_dbus_register_peer(wpa_s, dev_addr);
	}

	/* Notify a new peer has been detected*/
	wpas_dbus_signal_peer_device_found(wpa_s, dev_addr);
}


void wpas_notify_p2p_device_lost(struct wpa_supplicant *wpa_s,
				 const u8 *dev_addr)
{
	wpas_dbus_unregister_peer(wpa_s, dev_addr);

	/* Create signal on interface object*/
	wpas_dbus_signal_peer_device_lost(wpa_s, dev_addr);
}


void wpas_notify_p2p_group_removed(struct wpa_supplicant *wpa_s,
				   const struct wpa_ssid *ssid,
				   const char *role)
{
	wpas_dbus_signal_p2p_group_removed(wpa_s, role);

	wpas_dbus_unregister_p2p_group(wpa_s, ssid);
}


void wpas_notify_p2p_go_neg_req(struct wpa_supplicant *wpa_s,
				const u8 *src, u16 dev_passwd_id, u8 go_intent)
{
	wpas_dbus_signal_p2p_go_neg_req(wpa_s, src, dev_passwd_id, go_intent);
}


void wpas_notify_p2p_go_neg_completed(struct wpa_supplicant *wpa_s,
				      struct p2p_go_neg_results *res)
{
	wpas_dbus_signal_p2p_go_neg_resp(wpa_s, res);
}


void wpas_notify_p2p_invitation_result(struct wpa_supplicant *wpa_s,
				       int status, const u8 *bssid)
{
	wpas_dbus_signal_p2p_invitation_result(wpa_s, status, bssid);
}


void wpas_notify_p2p_sd_request(struct wpa_supplicant *wpa_s,
				int freq, const u8 *sa, u8 dialog_token,
				u16 update_indic, const u8 *tlvs,
				size_t tlvs_len)
{
	wpas_dbus_signal_p2p_sd_request(wpa_s, freq, sa, dialog_token,
					update_indic, tlvs, tlvs_len);
}


void wpas_notify_p2p_sd_response(struct wpa_supplicant *wpa_s,
				 const u8 *sa, u16 update_indic,
				 const u8 *tlvs, size_t tlvs_len)
{
	wpas_dbus_signal_p2p_sd_response(wpa_s, sa, update_indic,
					 tlvs, tlvs_len);
}


/**
 * wpas_notify_p2p_provision_discovery - Notification of provision discovery
 * @dev_addr: Who sent the request or responded to our request.
 * @request: Will be 1 if request, 0 for response.
 * @status: Valid only in case of response (0 in case of success)
 * @config_methods: WPS config methods
 * @generated_pin: PIN to be displayed in case of WPS_CONFIG_DISPLAY method
 *
 * This can be used to notify:
 * - Requests or responses
 * - Various config methods
 * - Failure condition in case of response
 */
void wpas_notify_p2p_provision_discovery(struct wpa_supplicant *wpa_s,
					 const u8 *dev_addr, int request,
					 enum p2p_prov_disc_status status,
					 u16 config_methods,
					 unsigned int generated_pin)
{
	wpas_dbus_signal_p2p_provision_discovery(wpa_s, dev_addr, request,
						 status, config_methods,
						 generated_pin);
}


void wpas_notify_p2p_group_started(struct wpa_supplicant *wpa_s,
				   struct wpa_ssid *ssid, int persistent,
				   int client, const u8 *ip)
{
	/* Notify a group has been started */
	wpas_dbus_register_p2p_group(wpa_s, ssid);

	wpas_dbus_signal_p2p_group_started(wpa_s, client, persistent, ip);
}


void wpas_notify_p2p_group_formation_failure(struct wpa_supplicant *wpa_s,
					     const char *reason)
{
	/* Notify a group formation failed */
	wpas_dbus_signal_p2p_group_formation_failure(wpa_s, reason);
}


void wpas_notify_p2p_wps_failed(struct wpa_supplicant *wpa_s,
				struct wps_event_fail *fail)
{
	wpas_dbus_signal_p2p_wps_failed(wpa_s, fail);
}


void wpas_notify_p2p_invitation_received(struct wpa_supplicant *wpa_s,
					 const u8 *sa, const u8 *go_dev_addr,
					 const u8 *bssid, int id, int op_freq)
{
	/* Notify a P2P Invitation Request */
	wpas_dbus_signal_p2p_invitation_received(wpa_s, sa, go_dev_addr, bssid,
						 id, op_freq);
}

void wpas_notify_p2p_bootstrap_req(struct wpa_supplicant *wpa_s,
				   const u8 *src, u16 bootstrap_method)
{
	wpas_dbus_signal_p2p_bootstrap_req(wpa_s, src, bootstrap_method);
}

void wpas_notify_p2p_bootstrap_rsp(struct wpa_supplicant *wpa_s,
				   const u8 *src, int status,
				   u16 bootstrap_method)
{
	wpas_dbus_signal_p2p_bootstrap_rsp(wpa_s, src, status,
					   bootstrap_method);
}

#endif /* CONFIG_P2P */


static void wpas_notify_ap_sta_authorized(struct wpa_supplicant *wpa_s,
					  const u8 *sta,
					  const u8 *p2p_dev_addr, const u8 *ip)
{
#ifdef CONFIG_P2P
	wpas_p2p_notify_ap_sta_authorized(wpa_s, p2p_dev_addr);

	/*
	 * Create 'peer-joined' signal on group object -- will also
	 * check P2P itself.
	 */
	if (p2p_dev_addr)
		wpas_dbus_signal_p2p_peer_joined(wpa_s, p2p_dev_addr);
#endif /* CONFIG_P2P */

	/* Register the station */
	wpas_dbus_register_sta(wpa_s, sta);

	/* Notify listeners a new station has been authorized */
	wpas_dbus_signal_sta_authorized(wpa_s, sta);
}


static void wpas_notify_ap_sta_deauthorized(struct wpa_supplicant *wpa_s,
					    const u8 *sta,
					    const u8 *p2p_dev_addr)
{
#ifdef CONFIG_P2P
	/*
	 * Create 'peer-disconnected' signal on group object if this
	 * is a P2P group.
	 */
	if (p2p_dev_addr)
		wpas_dbus_signal_p2p_peer_disconnected(wpa_s, p2p_dev_addr);
#endif /* CONFIG_P2P */

	/* Notify listeners a station has been deauthorized */
	wpas_dbus_signal_sta_deauthorized(wpa_s, sta);

	/* Unregister the station */
	wpas_dbus_unregister_sta(wpa_s, sta);
}


void wpas_notify_sta_authorized(struct wpa_supplicant *wpa_s,
				const u8 *mac_addr, int authorized,
				const u8 *p2p_dev_addr, const u8 *ip)
{
	if (authorized)
		wpas_notify_ap_sta_authorized(wpa_s, mac_addr, p2p_dev_addr,
					      ip);
	else
		wpas_notify_ap_sta_deauthorized(wpa_s, mac_addr, p2p_dev_addr);
}


void wpas_notify_certification(struct wpa_supplicant *wpa_s,
			       struct tls_cert_data *cert,
			       const char *cert_hash)
{
	int i;

	wpa_msg(wpa_s, MSG_INFO, WPA_EVENT_EAP_PEER_CERT
		"depth=%d subject='%s'%s%s%s%s",
		cert->depth, cert->subject, cert_hash ? " hash=" : "",
		cert_hash ? cert_hash : "",
		cert->tod == 2 ? " tod=2" : "",
		cert->tod == 1 ? " tod=1" : "");

	if (cert->cert) {
		char *cert_hex;
		size_t len = wpabuf_len(cert->cert) * 2 + 1;
		cert_hex = os_malloc(len);
		if (cert_hex) {
			wpa_snprintf_hex(cert_hex, len, wpabuf_head(cert->cert),
					 wpabuf_len(cert->cert));
			wpa_msg_ctrl(wpa_s, MSG_INFO,
				     WPA_EVENT_EAP_PEER_CERT
				     "depth=%d subject='%s' cert=%s",
				     cert->depth, cert->subject, cert_hex);
			os_free(cert_hex);
		}
	}

	for (i = 0; i < cert->num_altsubject; i++)
		wpa_msg(wpa_s, MSG_INFO, WPA_EVENT_EAP_PEER_ALT
			"depth=%d %s", cert->depth, cert->altsubject[i]);

	/* notify the new DBus API */
	wpas_dbus_signal_certification(wpa_s, cert->depth, cert->subject,
				       cert->altsubject, cert->num_altsubject,
				       cert_hash, cert->cert);
}


void wpas_notify_preq(struct wpa_supplicant *wpa_s,
		      const u8 *addr, const u8 *dst, const u8 *bssid,
		      const u8 *ie, size_t ie_len, u32 ssi_signal)
{
#ifdef CONFIG_AP
	wpas_dbus_signal_preq(wpa_s, addr, dst, bssid, ie, ie_len, ssi_signal);
#endif /* CONFIG_AP */
}


void wpas_notify_eap_status(struct wpa_supplicant *wpa_s, const char *status,
			    const char *parameter)
{
	wpas_dbus_signal_eap_status(wpa_s, status, parameter);
	wpa_msg_ctrl(wpa_s, MSG_INFO, WPA_EVENT_EAP_STATUS
		     "status='%s' parameter='%s'",
		     status, parameter);
}


void wpas_notify_eap_error(struct wpa_supplicant *wpa_s, int error_code)
{
	wpa_msg(wpa_s, MSG_ERROR, WPA_EVENT_EAP_ERROR_CODE "%d", error_code);
}


void wpas_notify_psk_mismatch(struct wpa_supplicant *wpa_s)
{
	wpas_dbus_signal_psk_mismatch(wpa_s);
}


void wpas_notify_sae_password_mismatch(struct wpa_supplicant *wpa_s)
{
	wpas_dbus_signal_sae_password_mismatch(wpa_s);
}


void wpas_notify_network_bssid_set_changed(struct wpa_supplicant *wpa_s,
					   struct wpa_ssid *ssid)
{
	if (wpa_s->current_ssid != ssid)
		return;

	wpa_dbg(wpa_s, MSG_DEBUG,
		"Network bssid config changed for the current network - within-ESS roaming %s",
		ssid->bssid_set ? "disabled" : "enabled");

	wpa_drv_roaming(wpa_s, !ssid->bssid_set,
			ssid->bssid_set ? ssid->bssid : NULL);
}


void wpas_notify_network_type_changed(struct wpa_supplicant *wpa_s,
				      struct wpa_ssid *ssid)
{
#ifdef CONFIG_P2P
	if (ssid->disabled == 2) {
		/* Changed from normal network profile to persistent group */
		ssid->disabled = 0;
		wpas_dbus_unregister_network(wpa_s, ssid->id);
		ssid->disabled = 2;
		ssid->p2p_persistent_group = 1;
		wpas_dbus_register_persistent_group(wpa_s, ssid);
	} else {
		/* Changed from persistent group to normal network profile */
		wpas_dbus_unregister_persistent_group(wpa_s, ssid->id);
		ssid->p2p_persistent_group = 0;
		wpas_dbus_register_network(wpa_s, ssid);
	}
#endif /* CONFIG_P2P */
}


#ifdef CONFIG_MESH

void wpas_notify_mesh_group_started(struct wpa_supplicant *wpa_s,
				    struct wpa_ssid *ssid)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_mesh_group_started(wpa_s, ssid);
}


void wpas_notify_mesh_group_removed(struct wpa_supplicant *wpa_s,
				    const u8 *meshid, u8 meshid_len,
				    u16 reason_code)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpas_dbus_signal_mesh_group_removed(wpa_s, meshid, meshid_len,
					    reason_code);
}


void wpas_notify_mesh_peer_connected(struct wpa_supplicant *wpa_s,
				     const u8 *peer_addr)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpa_msg(wpa_s, MSG_INFO, MESH_PEER_CONNECTED MACSTR,
		MAC2STR(peer_addr));
	wpas_dbus_signal_mesh_peer_connected(wpa_s, peer_addr);
}


void wpas_notify_mesh_peer_disconnected(struct wpa_supplicant *wpa_s,
					const u8 *peer_addr, u16 reason_code)
{
	if (wpa_s->p2p_mgmt)
		return;

	wpa_msg(wpa_s, MSG_INFO, MESH_PEER_DISCONNECTED MACSTR,
		MAC2STR(peer_addr));
	wpas_dbus_signal_mesh_peer_disconnected(wpa_s, peer_addr, reason_code);
}

#endif /* CONFIG_MESH */


#ifdef CONFIG_INTERWORKING

void wpas_notify_interworking_ap_added(struct wpa_supplicant *wpa_s,
				       struct wpa_bss *bss,
				       struct wpa_cred *cred, int excluded,
				       const char *type, int bh, int bss_load,
				       int conn_capab)
{
	wpa_msg(wpa_s, MSG_INFO, "%s" MACSTR " type=%s%s%s%s id=%d priority=%d sp_priority=%d",
		excluded ? INTERWORKING_EXCLUDED : INTERWORKING_AP,
		MAC2STR(bss->bssid), type,
		bh ? " below_min_backhaul=1" : "",
		bss_load ? " over_max_bss_load=1" : "",
		conn_capab ? " conn_capab_missing=1" : "",
		cred->id, cred->priority, cred->sp_priority);

	wpas_dbus_signal_interworking_ap_added(wpa_s, bss, cred, type, excluded,
					       bh, bss_load, conn_capab);
}


void wpas_notify_interworking_select_done(struct wpa_supplicant *wpa_s)
{
	wpas_dbus_signal_interworking_select_done(wpa_s);
}


void wpas_notify_anqp_query_done(struct wpa_supplicant *wpa_s,
				 const u8 *dst, const char *result)
{
	wpa_msg(wpa_s, MSG_INFO, ANQP_QUERY_DONE "addr=" MACSTR " result=%s",
		MAC2STR(dst), result);

	wpas_dbus_signal_anqp_query_done(wpa_s, dst, result);
}

#endif /* CONFIG_INTERWORKING */


void wpas_notify_pmk_cache_added(struct wpa_supplicant *wpa_s,
				 struct rsn_pmksa_cache_entry *entry)
{
	/* TODO: Notify external entities of the added PMKSA cache entry */
}


void wpas_notify_signal_change(struct wpa_supplicant *wpa_s)
{
	wpas_dbus_signal_prop_changed(wpa_s, WPAS_DBUS_PROP_SIGNAL_CHANGE);
}


#ifdef CONFIG_HS20
void wpas_notify_hs20_t_c_acceptance(struct wpa_supplicant *wpa_s,
				     const char *url)
{
	wpa_msg(wpa_s, MSG_INFO, HS20_T_C_ACCEPTANCE "%s", url);
	wpas_dbus_signal_hs20_t_c_acceptance(wpa_s, url);
}
#endif /* CONFIG_HS20 */


#if defined(CONFIG_NAN) || defined(CONFIG_NAN_USD)

void wpas_notify_nan_discovery_result(struct wpa_supplicant *wpa_s,
				      struct nan_discovery_result *res)
{
	char *ssi_hex, *pmkid_hex = NULL;
	char *cipher_suites_str = NULL;

	ssi_hex = os_zalloc(2 * res->ssi_len + 1);
	if (!ssi_hex)
		return;
	if (res->ssi)
		wpa_snprintf_hex(ssi_hex, 2 * res->ssi_len + 1,
				 res->ssi, res->ssi_len);

	if (res->pmkid_list && res->pmkid_count > 0) {
		const size_t pmkid_hex_len = 2 * PMKID_LEN + 1;
		unsigned int i;

		pmkid_hex = os_zalloc(res->pmkid_count * pmkid_hex_len);
		if (pmkid_hex) {
			for (i = 0; i < res->pmkid_count; i++) {
				char *pos = &pmkid_hex[i * pmkid_hex_len];

				wpa_snprintf_hex(
					pos, pmkid_hex_len,
					&res->pmkid_list[i * PMKID_LEN],
					PMKID_LEN);
				if (i < res->pmkid_count - 1)
					pos[2 * PMKID_LEN] = ',';
			}
		}
	}

	if (res->cipher_suites && res->n_cipher_suites > 0) {
		char *pos;
		unsigned int i;

		/* Allocate enough space for trailing space after each cipher */
		cipher_suites_str = os_zalloc(res->n_cipher_suites * 2);
		if (!cipher_suites_str)
			goto err;

		pos = cipher_suites_str;

		for (i = 0; i < res->n_cipher_suites; i++) {
			int ret;
			size_t left = res->n_cipher_suites * 2 -
				(pos - cipher_suites_str);

			ret = os_snprintf(pos, left, "%s%u", i > 0 ? "," : "",
					  res->cipher_suites[i]);
			if (os_snprintf_error(left, ret))
				break;
			pos += ret;
		}
	}
err:

	wpa_msg_global(wpa_s, MSG_INFO, NAN_DISCOVERY_RESULT
		       "subscribe_id=%d publish_id=%d address=" MACSTR
		       " fsd=%d fsd_gas=%d srv_proto_type=%u ssi=%s%s%s%s%s pairing_setup_supp=%d npk_nik_caching_supp=%d pbm=0x%04x data_path=%d",
		       res->subscribe_id, res->peer_publish_id,
		       MAC2STR(res->peer_addr),
		       res->fsd, res->fsd_gas, res->srv_proto_type, ssi_hex,
		       pmkid_hex ? " pmkid=" : "",
		       pmkid_hex ? pmkid_hex : "",
		       cipher_suites_str ? " cipher_suites=" : "",
		       cipher_suites_str ? cipher_suites_str : "",
		       res->pairing_setup_supp, res->npk_nik_caching_supp,
		       res->pbm, res->data_path);
	os_free(ssi_hex);
	os_free(pmkid_hex);
	os_free(cipher_suites_str);

	wpas_dbus_signal_nan_discovery_result(wpa_s,
					      res->srv_proto_type,
					      res->subscribe_id,
					      res->peer_publish_id,
					      res->peer_addr,
					      res->fsd, res->fsd_gas,
					      res->ssi, res->ssi_len);
}


void wpas_notify_nan_replied(struct wpa_supplicant *wpa_s,
			     enum nan_service_protocol_type srv_proto_type,
			     int publish_id, int peer_subscribe_id,
			     const u8 *peer_addr,
			     const u8 *ssi, size_t ssi_len)
{
	char *ssi_hex;

	ssi_hex = os_zalloc(2 * ssi_len + 1);
	if (!ssi_hex)
		return;
	if (ssi)
		wpa_snprintf_hex(ssi_hex, 2 * ssi_len + 1, ssi, ssi_len);
	wpa_msg_global(wpa_s, MSG_INFO, NAN_REPLIED
		       "publish_id=%d address=" MACSTR
		       " subscribe_id=%d srv_proto_type=%u ssi=%s",
		       publish_id, MAC2STR(peer_addr), peer_subscribe_id,
		       srv_proto_type, ssi_hex);
	os_free(ssi_hex);

	wpas_dbus_signal_nan_replied(wpa_s, srv_proto_type, publish_id,
				     peer_subscribe_id, peer_addr,
				     ssi, ssi_len);
}


void wpas_notify_nan_receive(struct wpa_supplicant *wpa_s, int id,
			     int peer_instance_id, const u8 *peer_addr,
			     const u8 *ssi, size_t ssi_len)
{
	char *ssi_hex;

	ssi_hex = os_zalloc(2 * ssi_len + 1);
	if (!ssi_hex)
		return;
	if (ssi)
		wpa_snprintf_hex(ssi_hex, 2 * ssi_len + 1, ssi, ssi_len);
	wpa_msg_global(wpa_s, MSG_INFO, NAN_RECEIVE
		       "id=%d peer_instance_id=%d address=" MACSTR " ssi=%s",
		       id, peer_instance_id, MAC2STR(peer_addr), ssi_hex);
	os_free(ssi_hex);

	wpas_dbus_signal_nan_receive(wpa_s, id, peer_instance_id, peer_addr,
				     ssi, ssi_len);
}


static const char * nan_reason_txt(enum nan_de_reason reason)
{
	switch (reason) {
	case NAN_DE_REASON_TIMEOUT:
		return "timeout";
	case NAN_DE_REASON_USER_REQUEST:
		return "user-request";
	case NAN_DE_REASON_FAILURE:
		return "failure";
	}

	return "unknown";
}


void wpas_notify_nan_publish_terminated(struct wpa_supplicant *wpa_s,
					int publish_id,
					enum nan_de_reason reason)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_PUBLISH_TERMINATED
		       "publish_id=%d reason=%s",
		       publish_id, nan_reason_txt(reason));
	wpas_dbus_signal_nan_publish_terminated(wpa_s, publish_id,
						nan_reason_txt(reason));
}


void wpas_notify_nan_subscribe_terminated(struct wpa_supplicant *wpa_s,
					  int subscribe_id,
					  enum nan_de_reason reason)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_SUBSCRIBE_TERMINATED
		       "subscribe_id=%d reason=%s",
		       subscribe_id, nan_reason_txt(reason));
	wpas_dbus_signal_nan_subscribe_terminated(wpa_s, subscribe_id,
						  nan_reason_txt(reason));
}


void wpas_notify_nan_transmit_req_status(struct wpa_supplicant *wpa_s,
					 u32 cookie, bool acked)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_TRANSMIT_STATUS
		       "cookie=%u acked=%u", cookie, acked);
}


void wpas_notify_nan_bootstrap_request(struct wpa_supplicant *wpa_s,
				       const u8 *peer_nmi, u16 pbm,
				       int handle, u8 requestor_instance_id)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_BOOTSTRAP_REQUEST
		       "peer_nmi=" MACSTR
		       " pbm=0x%04x handle=%d requestor_instance_id=%u",
		       MAC2STR(peer_nmi), pbm, handle, requestor_instance_id);
}


void wpas_notify_nan_bootstrap_success(struct wpa_supplicant *wpa_s,
				       const u8 *peer_nmi, u16 pbm,
				       int handle, u8 requestor_instance_id)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_BOOTSTRAP_SUCCESS
		       "peer_nmi=" MACSTR
		       " pbm=0x%04x handle=%d requestor_instance_id=%u",
		       MAC2STR(peer_nmi), pbm, handle, requestor_instance_id);
}


void wpas_notify_nan_bootstrap_failure(struct wpa_supplicant *wpa_s,
				       const u8 *peer_nmi, u16 pbm,
				       u8 reason, int handle,
				       u8 requestor_instance_id)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_BOOTSTRAP_FAILURE
		       "peer_nmi=" MACSTR
		       " pbm=0x%04x reason=%u handle=%d requestor_instance_id=%u",
		       MAC2STR(peer_nmi), pbm, reason, handle,
		       requestor_instance_id);
}


#ifdef CONFIG_PASN

void wpas_notify_nan_nik_received(struct wpa_supplicant *wpa_s,
				  const u8 *nik, size_t nik_len,
				  int cipher_ver, int akmp,
				  const u8 *npk, size_t npk_len,
				  int nik_lifetime, int identity_id)
{
	char *nik_hex, *npk_hex;
	size_t nik_hex_len = 2 * nik_len + 1;
	size_t npk_hex_len = 2 * npk_len + 1;

	nik_hex = os_malloc(nik_hex_len);
	npk_hex = os_malloc(npk_hex_len);
	if (!nik_hex || !npk_hex) {
		wpa_printf(MSG_INFO,
			   "NAN NIK received: Failed to allocate memory");
		goto out;
	}

	wpa_snprintf_hex(nik_hex, nik_hex_len, nik, nik_len);
	wpa_snprintf_hex(npk_hex, npk_hex_len, npk, npk_len);

	wpa_msg_global(wpa_s, MSG_INFO, NAN_NIK_RECEIVED
		       "nik=%s cipher_version=%d akmp=%s pmk=%s lifetime=%d identity_id=%d",
		       nik_hex, cipher_ver,
		       wpa_key_mgmt_txt(akmp, WPA_PROTO_RSN),
		       npk_hex, nik_lifetime, identity_id);

out:
	bin_clear_free(nik_hex, nik_hex_len);
	bin_clear_free(npk_hex, npk_hex_len);
}


void wpas_notify_nan_pairing_request(struct wpa_supplicant *wpa_s,
				     const u8 *peer_nmi, u8 csid,
				     u8 instance_id, int key_mgmt,
				     bool verify)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_PAIRING_REQUEST
		       "peer_nmi=" MACSTR
		       " csid=%u instance_id=%u key_mgmt=%s verify=%d",
		       MAC2STR(peer_nmi), csid, instance_id,
		       wpa_key_mgmt_txt(key_mgmt, WPA_PROTO_RSN),
		       verify);
}

#endif /* CONFIG_PASN */


void wpas_notify_nan_ndp_request(struct wpa_supplicant *wpa_s,
				 const u8 *peer_nmi, const u8 *init_ndi,
				 u32 ndp_id, u8 publish_inst_id,
				 const u8 *ssi, size_t ssi_len, u32 csid)
{
	char *ssi_hex = NULL;

	if (ssi) {
		ssi_hex = os_zalloc(2 * ssi_len + 1);
		if (!ssi_hex)
			return;

		wpa_snprintf_hex(ssi_hex, 2 * ssi_len + 1, ssi, ssi_len);
	}

	wpa_msg_global(wpa_s, MSG_INFO, NAN_NDP_REQUEST "peer_nmi=" MACSTR
		       " init_ndi=" MACSTR
		       " ndp_id=%u publish_inst_id=%u ssi=%s csid=%u",
		       MAC2STR(peer_nmi), MAC2STR(init_ndi),
		       ndp_id, publish_inst_id,
		       ssi_hex ? ssi_hex : "", csid);

	os_free(ssi_hex);
}


void wpas_notify_nan_ndp_counter_request(struct wpa_supplicant *wpa_s,
					 const u8 *peer_nmi,
					 const u8 *init_ndi, u32 ndp_id,
					 const u8 *ssi, size_t ssi_len)
{
	char *ssi_hex = NULL;

	if (ssi) {
		ssi_hex = os_zalloc(2 * ssi_len + 1);
		if (!ssi_hex)
			return;

		wpa_snprintf_hex(ssi_hex, 2 * ssi_len + 1, ssi, ssi_len);
	}

	wpa_msg_global(wpa_s, MSG_INFO, NAN_NDP_COUNTER_REQUEST
		       "peer_nmi=" MACSTR " init_ndi=" MACSTR
		       " ndp_id=%u ssi=%s",
		       MAC2STR(peer_nmi), MAC2STR(init_ndi),
		       ndp_id, ssi_hex ? ssi_hex : "");

	os_free(ssi_hex);
}


void wpas_notify_nan_ndp_connected(struct wpa_supplicant *wpa_s,
				   const u8 *peer_nmi, u32 ndp_id,
				   const u8 *local_ndi,
				   const u8 *peer_ndi,
				   const u8 *ssi, size_t ssi_len,
				   const u8 *interface_id)
{
	char *ssi_hex = NULL, *interface_id_hex = NULL;

	if (ssi) {
		ssi_hex = os_zalloc(2 * ssi_len + 1);
		if (!ssi_hex)
			return;

		wpa_snprintf_hex(ssi_hex, 2 * ssi_len + 1, ssi, ssi_len);
	}

	if (interface_id) {
		interface_id_hex =
			os_zalloc(2 * NAN_NDPE_TLV_IPV6_LINK_LOCAL_LEN + 1);
		if (!interface_id_hex)
			return;

		wpa_snprintf_hex(interface_id_hex,
				 2 * NAN_NDPE_TLV_IPV6_LINK_LOCAL_LEN + 1,
				 interface_id,
				 NAN_NDPE_TLV_IPV6_LINK_LOCAL_LEN);
	}

	wpa_msg_global(wpa_s, MSG_INFO, NAN_NDP_CONNECTED "peer=" MACSTR
		       " ndp_id=%u local_ndi=" MACSTR
		       " peer_ndi=" MACSTR " ssi=%s interface_id=%s",
		       MAC2STR(peer_nmi), ndp_id,
		       MAC2STR(local_ndi), MAC2STR(peer_ndi),
		       ssi_hex ? ssi_hex : "",
		       interface_id ? interface_id_hex : "");

	os_free(ssi_hex);
	os_free(interface_id_hex);
}


void wpas_notify_nan_ndp_disconnected(struct wpa_supplicant *wpa_s,
				      const u8 *peer_nmi, u32 ndp_id,
				      const u8 *local_ndi,
				      const u8 *peer_ndi,
				      enum nan_reason reason,
				      bool locally_generated,
				      bool failure)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_NDP_DISCONNECTED
		       "peer=" MACSTR " ndp_id=%u local_ndi=" MACSTR
		       " peer_ndi=" MACSTR
		       " reason=%u locally_generated=%d, failure=%d",
		       MAC2STR(peer_nmi), ndp_id,
		       MAC2STR(local_ndi), MAC2STR(peer_ndi), reason,
		       locally_generated, failure);
}


void wpas_notify_nan_cluster_join(struct wpa_supplicant *wpa_s,
				  const u8 *cluster_id, bool new_cluster)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_CLUSTER_JOIN "cluster_id=" MACSTR
		       " new=%d", MAC2STR(cluster_id), new_cluster);
}


void wpas_notify_nan_schedule_changed(struct wpa_supplicant *wpa_s,
				      const u8 *peer_nmi)
{
	wpa_msg_global(wpa_s, MSG_INFO,
		       NAN_PEER_SCHEDULE_CHANGED "peer_nmi=" MACSTR,
		       MAC2STR(peer_nmi));
}


void wpas_notify_nan_sched_update_done(struct wpa_supplicant *wpa_s,
				       bool success)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_SCHEDULE_UPDATE_DONE "status=%s",
		       success ? "success" : "failure");
}


void wpas_notify_nan_pairing_status(struct wpa_supplicant *wpa_s,
				    const u8 *peer_addr, int akmp,
				    int cipher, u16 status,
				    const u8 *nd_pmk)
{
	char nd_pmk_hex[2 * PMK_LEN + 1];

	if (nd_pmk)
		wpa_snprintf_hex(nd_pmk_hex, sizeof(nd_pmk_hex), nd_pmk,
				 PMK_LEN);
	else
		nd_pmk_hex[0] = '\0';

	wpa_msg_global(wpa_s, MSG_INFO, NAN_PAIRING_STATUS "addr=" MACSTR
		       " akmp=%s cipher=%s status=%s%s%s",
		       MAC2STR(peer_addr),
		       wpa_key_mgmt_txt(akmp, WPA_PROTO_RSN),
		       wpa_cipher_txt(cipher),
		       status == WLAN_STATUS_SUCCESS ? "success" : "failure",
		       nd_pmk ? " nd_pmk=" : "",
		       nd_pmk ? nd_pmk_hex : "");
	forced_memzero(nd_pmk_hex, sizeof(nd_pmk_hex));
}


void wpas_notify_nan_chan_evacuation(struct wpa_supplicant *wpa_s,
				     u8 map_id, int freq)
{
	wpa_msg_global(wpa_s, MSG_INFO,
		       NAN_CHAN_EVACUATION "map_id=%u freq=%d",
		       map_id, freq);
}


void wpas_notify_nan_stopped(struct wpa_supplicant *wpa_s)
{
	wpa_msg_global(wpa_s, MSG_INFO, NAN_STOPPED "ifname=%s",
		       wpa_s->ifname);
}

#endif /* CONFIG_NAN || CONFIG_NAN_USD */


#ifdef CONFIG_PR

void wpas_notify_pr_pasn_result(struct wpa_supplicant *wpa_s, u8 role,
				u8 protocol_type, u8 op_class, u8 op_channel,
				const char *country)
{
	wpa_msg_global(wpa_s, MSG_INFO, PR_PASN_RESULT
		       "SUCCESS role=%u protocol=%u opclass=%u channel=%u cc=%c%c",
		       role, protocol_type, op_class, op_channel,
		       country[0], country[1]);
}


void wpas_notify_pr_negotiation_started(struct wpa_supplicant *wpa_s,
					const u8 *peer_addr, u8 role,
					u8 protocol_type)
{
	wpa_msg_global(wpa_s, MSG_INFO, PR_PASN_NEGOTIATION_STARTED
		       "peer_addr=" MACSTR " role=%u protocol=%u",
		       MAC2STR(peer_addr), role, protocol_type);
}


void wpas_notify_pr_ranging_params(struct wpa_supplicant *wpa_s,
				   const u8 *dev_addr, const u8 *peer_addr,
				   u8 ranging_role, u8 protocol_type, int freq,
				   int channel, int bw, int format_bw)
{
	wpa_msg_global(wpa_s, MSG_INFO, PR_RANGING_PARAMS
		       "dev_addr=" MACSTR " peer_addr=" MACSTR
		       " role=%u protocol=%u freq=%d channel=%d bw=%d format_bw=%d",
		       MAC2STR(dev_addr), MAC2STR(peer_addr), ranging_role,
		       protocol_type, freq, channel, bw, format_bw);
}


void wpas_notify_pr_measurement_result(
	struct wpa_supplicant *wpa_s,
	const struct peer_measurement_result *result)
{
	char rtt[32] = "";
	char dist[32] = "";

	if (result->ftm.fail) {
		wpa_msg_global(wpa_s, MSG_INFO, PR_EVENT_PEER_MEASUREMENT
			       "addr=" MACSTR
			       " status=%u burst_index=%u fail=1 fail_reason=%u",
			       MAC2STR(result->addr), result->status,
			       result->ftm.burst_index,
			       result->ftm.fail_reason);
		return;
	}

	if (result->ftm.rtt_avg || result->ftm.has_data)
		os_snprintf(rtt, sizeof(rtt), " rtt_avg=%lld",
			    (long long) result->ftm.rtt_avg);
	if (result->ftm.dist_avg || result->ftm.has_data)
		os_snprintf(dist, sizeof(dist), " dist_avg=%lld",
			    (long long) result->ftm.dist_avg);

	wpa_msg_global(wpa_s, MSG_INFO, PR_EVENT_PEER_MEASUREMENT
		       "addr=" MACSTR " status=%u burst_index=%u%s%s",
		       MAC2STR(result->addr), result->status,
		       result->ftm.burst_index, rtt, dist);
}


void wpas_notify_pr_ranging_complete(struct wpa_supplicant *wpa_s, u64 cookie)
{
	wpa_msg_global(wpa_s, MSG_INFO, PR_EVENT_RANGING_COMPLETE
		       "cookie=%llu", (unsigned long long) cookie);
}


void wpas_notify_pr_device_found(struct wpa_supplicant *wpa_s,
				 const struct pr_device *dev)
{
	wpa_msg_global(wpa_s, MSG_INFO, PR_PEER_FOUND
		       "peer_addr=" MACSTR
		       " pasn_type=0x%02x name=%s edca=%d edca_ista=%d edca_rsta=%d ntb=%d ntb_ista=%d ntb_rsta=%d secure_ltf=%d 6ghz=%d freq=%d dik_valid=%d",
		       MAC2STR(dev->pr_device_addr),
		       dev->pr_caps.pasn_type,
		       dev->pr_caps.device_name,
		       dev->pr_caps.edca_support,
		       dev->edca_caps.ista_support,
		       dev->edca_caps.rsta_support,
		       dev->pr_caps.ntb_support,
		       dev->ntb_caps.ista_support,
		       dev->ntb_caps.rsta_support,
		       dev->pr_caps.secure_he_ltf,
		       dev->pr_caps.support_6ghz,
		       dev->listen_freq,
		       dev->dik_valid);
}


void wpas_notify_pr_ranging_terminated(struct wpa_supplicant *wpa_s, int reason)
{
	wpa_msg_global(wpa_s, MSG_INFO, PR_RANGING_TERMINATED "reason=%d",
		       reason);
}

#endif /* CONFIG_PR */
